import os
import time
import torch

from typing import List, Dict, Any
from typing import TypedDict, Annotated
from langgraph.graph import StateGraph, START, END
from langgraph.graph.message import add_messages
from langchain_openai.chat_models import ChatOpenAI
from langchain.embeddings import HuggingFaceEmbeddings
from langchain_community.vectorstores import FAISS
from langchain_core.messages import BaseMessage
from collections import deque
from tool_collection.argo.ARGO import ArgoWrapper
from prompts import query_prompt, search_prompt, summarize_prompt, critic_prompt, evaluate_prompt


os.environ["OPENAI_API_KEY"] = "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"
FAISS_INDEX_PATH="../../FAISS-all-mpnet-base-v2-nougat/index.faiss"
SENTENCE_TRANSFORMER_MODEL_NAME="sentence-transformers/all-mpnet-base-v2"

def get_faiss_storage_alphavirus():
    embedding_model = HuggingFaceEmbeddings(model_name=SENTENCE_TRANSFORMER_MODEL_NAME)
    index_path = FAISS_INDEX_PATH
    vector_store = FAISS.load_local(
        index_path,
        embeddings=embedding_model,
        allow_dangerous_deserialization=True
    )

    return vector_store


# Define the RAG tool
def rag_query(vector_sorage: Any,
             query: str, 
             number_of_retrieved_documents: int = 20) -> List[str]:
    """
    Retrieve relevant documents from the FAISS vector store based on the query.
    Args:
        query: The search query to find relevant documents
        number_of_retrieved_documents: The number of documents to retrieve from vector storage
    Returns:
        List of relevant document contents
    """
    docs = vector_sorage.similarity_search(query, k=number_of_retrieved_documents)
    return [doc.page_content for doc in docs]

# Define the state
class State(TypedDict):
    messages: Annotated[list[BaseMessage], add_messages]
    original_query: str
    current_query: str
    search_attempts: int
    critic_response: str
    retrieved_docs: List[str]
    summary: str
    previous_answer: str
    current_answer: str
    next_step: str


class RAGAgentTeam():
    def __init__(self, llm_name: str = "gpt-4", max_search_attempts=3):
        self.llm = ArgoWrapper() #ChatOpenAI(model=llm_name, temperature=0)
        self.vector_storage = get_faiss_storage_alphavirus()
        self.states_history = []
        self.max_search_attempts = max_search_attempts
        self.token_history = deque()
        self.tpm_limit = 60000

    def track_tokens(self, response):
        """Track token usage from an OpenAI API response"""
        if "usage" in response:
            tokens_used = response["usage"]["total_tokens"]
            self.token_history.append((time.time(), tokens_used))
        
        # Remove old entries beyond 1 minute
        while self.token_history and time.time() - self.token_history[0][0] > 60:
            self.token_history.popleft()
                
    def get_tpm_usage(self):
        """Calculate current tokens per minute usage"""
        return sum(tokens for _, tokens in self.token_history)
        
    def check_tpm_limit(self):
        """Check if the model is exceeding the TPM limit"""
        return self.get_tpm_usage() < self.tpm_limit

    def create_graph(self) -> StateGraph:
        #tools = [partial(self.search_tool, vector_sorage=self.vector_sorage)]
        #tool_node = ToolNode(tools)
        #model = self.llm.bind_tools(tools)

        workflow = StateGraph(State)
    
        # Add nodes
        workflow.add_node("search", self.search_agent)
        workflow.add_node("answer", self.answer_agent)
        workflow.add_node("evaluate", self.evaluation_agent)
    
        # Add edges
        workflow.add_edge("search", "answer")
        workflow.add_edge("answer", "evaluate")
    
        # Add conditional edges with router function
        workflow.add_conditional_edges(
            "evaluate",
            self.router,
            {
                "continue_search": "search",
                "end": END
            }
        )
    
        workflow.set_entry_point("search")
    
        return workflow.compile()
    

    def answer(self, state: State) -> State:
        state["current_answer"] = self.llm.invoke(
                query_prompt.format(
                    user_query=state["current_query"] if state["search_attempts"] > 0 and state["search_attempts"] < self.max_search_attempts-1 else state["original_query"],
                    criticism=state["critic_response"],
                    context=state["summary"]
                )
            )
    
        return state
    

    def extract_response(self, response):
        if type(self.llm) is ArgoWrapper:
            response = response['response']
        else:
            response = response.content
        
        return response
    

    def search(self, state: State) -> State:
        """Execute RAG search based on the current query"""
        if not state.get("search_attempts"):
            state["search_attempts"] = 0
        
        if state["search_attempts"] == 0:
            state["original_query"] = state["current_query"]


        #breakpoint()         
        # Generate search query
        search_response = self.llm.invoke(
            search_prompt.format(
                question=state["current_query"],
                previous_answer=state["previous_answer"],
                previous_summary=state.get("summary", ""),
                attempt_number=state["search_attempts"] + 1
            )
        )


        state["retrieved_docs"] = rag_query(self.vector_storage, 
                                            query=self.extract_response(search_response), #if state["search_attempts"] > 0 else state["original_query"],
                                            number_of_retrieved_documents=20)
        state["search_attempts"] += 1
        #self.states_history.append(state)
        return state

    def search_agent(self, state: State) -> State:
        state = self.search(state)
        return state

    def summarize(self, state: State) -> State:
        """Summarize retrieved documents"""

        summary_response = self.llm.invoke(
            summarize_prompt.format(
                question=state["original_query"],
                documents='\n\n'.join(state["retrieved_docs"])
            )
        )
        state["summary"] = self.extract_response(summary_response) #summary_response.content
        #self.states_history.append(state)
        return state

    def critic(self, state: State) -> State:
        """Critic a response"""

        critic_response = self.llm.invoke(
            critic_prompt.format(
                question=state["current_query"],
                answer=state["current_answer"]
            )
        )

        state["critic_response"] = critic_response
        return state

    def evaluate(self, state: State) -> State:
        """Evaluate if information is sufficient or if another search is needed"""

        eval_response = self.llm.invoke(
            evaluate_prompt.format(
                question=state["current_query"],
                answer=state["current_answer"],
                summary=state["summary"],
                critic_response=state["critic_response"],
                attempt_number=state["search_attempts"]
            )
        )

        decision, explanation = self.extract_response(eval_response).split(" | ", 1) # .content

        if decision == "SUFFICIENT":
            state["next_step"] = "end"
        elif state["search_attempts"] == 3:
            state["next_step"] = "end"
        else:
            state["next_step"] = "continue_search"
            state["previous_answer"] = state["current_answer"]

        return state


    def answer_agent(self, state: State) -> State:
        state = self.summarize(state)
        state = self.answer(state)
        return state

    def evaluation_agent(self, state: State) -> State:
        if not self.check_tpm_limit():
            time.sleep(60)

        state = self.critic(state)
        state = self.evaluate(state)
        return state
    

    def router(self, state: State) -> str:
        """Route to the next node based on the evaluation decision"""
        return state["next_step"]

    def run(self, user_input):
        graph = self.create_graph()

            
        # Initialize state
        state = State(
            messages=[],
            current_query=user_input,
            search_attempts=0,
            critic_response="",
            retrieved_docs=[],
            summary="",
            previous_answer="",
            current_answer="",
            next_step=""
        )
        
        # Run the graph
        final_state = graph.invoke(state)
        return final_state