import uvicorn

from fastapi import FastAPI, HTTPException
from pydantic import BaseModel
from agent_team import RAGAgentTeam



# Define request model
class Query(BaseModel):
    text: str
    rag_flag: bool
    num_docs: int = 1  # Default to 1 document if not specified

# Define response model
class Response(BaseModel):
    answer: str
    context: str
    

if __name__ == "__main__":
    # Initialize FastAPI app
    app = FastAPI(title="RAG API Server",
                description="API for querying LangGraph-based RAG system",
                version="1.0.0")
    app.agent = RAGAgentTeam(llm_name="gpt-4o")


    @app.post("/query", response_model=Response)
    async def process_query(query: Query):
        try:
            response = None
            if query.rag_flag:
                result_state = app.agent.run(query.text)        

                current_answer = app.agent.extract_response(result_state["current_answer"])
                summary = result_state["summary"]
                criticism = app.agent.extract_response(result_state["critic_response"])

                # Extract answer and retrieved documents
                response = Response(
                    answer='\n\n\n'.join(['**ANSWER**', current_answer, '**CONTEXT**', summary, '**CRITICAL ANALYSIS**', criticism]),
                    context=result_state["summary"]
                )
            else:
                result_state = app.agent.llm.invoke(query.text)
                response = Response(
                    answer = app.agent.extract_response(result_state),
                    context=''
                )
            return response
        except Exception as e:
            print(str(e))
            raise HTTPException(status_code=500, detail=str(e))

    @app.get("/health")
    async def health_check():
        return {"status": "healthy"}

    uvicorn.run(app, host="0.0.0.0", port=8121)
